﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;

namespace WpfAudioPlayer.Controls
{
	/// <summary>
	/// ホイールスクロールを常に有効にする為のユーティリティクラス。
	/// </summary>
	sealed class ScrollViewerCorrector
	{
		/// <summary>
		/// ホイールクロールを有効にするかを示す値を取得する。
		/// </summary>
		/// <param name="obj">依存オブジェクト。</param>
		/// <returns>ホイールクロールを有効にするかを示す値。</returns>
		public static bool GetFixScrolling( DependencyObject obj )
		{
			return ( bool )obj.GetValue( ScrollViewerCorrector.FixScrollingProperty );
		}

		/// <summary>
		/// ホイールクロールを有効にするかを示す値を設定する。
		/// </summary>
		/// <param name="obj">依存オブジェクト。</param>
		/// <param name="value">ホイールクロールを有効にする場合は treu、それ以外は false。</param>
		public static void SetFixScrolling( DependencyObject obj, bool value )
		{
			obj.SetValue( ScrollViewerCorrector.FixScrollingProperty, value );
		}

		/// <summary>
		/// ホイールクロールを有効にするかを示す値の依存プロパティ。
		/// </summary>
		public static readonly DependencyProperty FixScrollingProperty = DependencyProperty.RegisterAttached( "FixScrolling", typeof( bool ), typeof( ScrollViewerCorrector ), new FrameworkPropertyMetadata( false, ScrollViewerCorrector.OnFixScrollingPropertyChanged ) );

		/// <summary>
		/// ホイールクロールを有効にするかを示す値が変更された時のイベント。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベントデータ。</param>
		public static void OnFixScrollingPropertyChanged( object sender, DependencyPropertyChangedEventArgs e )
		{
			ScrollViewer viewer = sender as ScrollViewer;
			if( viewer == null ) { throw new ArgumentException( "The dependency property can only be attached to a ScrollViewer", "sender" ); }

			if( ( bool )e.NewValue == true )
			{
				viewer.PreviewMouseWheel += HandlePreviewMouseWheel;
			}
			else if( ( bool )e.NewValue == false )
			{
				viewer.PreviewMouseWheel -= HandlePreviewMouseWheel;
			}
		}

		/// <summary>
		/// マウスホイールを処理するイベントのコレクション。
		/// </summary>
		private static List< MouseWheelEventArgs > _reentrantList = new List< MouseWheelEventArgs >();
		
		/// <summary>
		/// マウスホイールが操作された時のイベント。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベントデータ。</param>
		private static void HandlePreviewMouseWheel( object sender, MouseWheelEventArgs e )
		{
			var scrollControl = sender as ScrollViewer;
			if( !e.Handled && sender != null && !_reentrantList.Contains( e ) )
			{
				var previewEventArg = new MouseWheelEventArgs( e.MouseDevice, e.Timestamp, e.Delta )
				{
					RoutedEvent = UIElement.PreviewMouseWheelEvent,
					Source = sender
				};
				
				var originalSource = e.OriginalSource as UIElement;
				ScrollViewerCorrector._reentrantList.Add( previewEventArg );
				originalSource.RaiseEvent( previewEventArg );
				ScrollViewerCorrector._reentrantList.Remove( previewEventArg );

				// at this point if no one else handled the event in our children, we do our job
				if( !previewEventArg.Handled && ( ( e.Delta > 0 && scrollControl.VerticalOffset == 0 )
					|| ( e.Delta <= 0 && scrollControl.VerticalOffset >= scrollControl.ExtentHeight - scrollControl.ViewportHeight ) ) )
				{
					e.Handled = true;
					
					var eventArg = new MouseWheelEventArgs( e.MouseDevice, e.Timestamp, e.Delta );
					eventArg.RoutedEvent = UIElement.MouseWheelEvent;
					eventArg.Source = sender;
					
					var parent = ( UIElement )( ( FrameworkElement )sender ).Parent;
					parent.RaiseEvent( eventArg );
				}
			}
		}
	}
}
